/*
 * dtrace.c - dtrace routines.
 *
 * $Id: dtrace.c 32 2006-05-31 15:28:13Z brendan $
 *
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 *
 * COPYRIGHT: Copyright (c) 2006 Brendan Gregg.
 *
 * 29-Jan-2006  Brendan Gregg   Created this
 */

#include <gtk/gtk.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/filio.h>
#include <signal.h>
#include <strings.h>

#include "dtrace.h"
#include "system.h"

/* public variables */
enum dt_state	dt_state = DT_STATE_STOPPED;	/* tracing on/off */
enum dt_script	dt_script = DT_SCRIPT_SINGLE;	/* tracing a single slice */
int		dt_heartbeat = 100;		/* heartbeat & switchrate ms */

/* private variables */
static FILE	*dt_dtrace_in;			/* FILE to dtrace coprocess */
static pid_t	dt_dtrace_pid;			/* dtrace PID */
static int	dt_dtrace_check = 0;		/* dtrace pulse counter */

/*
 * DTrace script, this gets merged later
 */
static const char 	*dt_script0 = "io:::done /args[0]->b_edev == ";
static const char 	*dt_script1 = " / {"
"	printf(\"%x\\n\", args[0]->b_blkno);"
"}"
"profile:::tick-";
static const char 	*dt_script2 = "ms { trace(\"h\\n\"); }";

static void
dtrace_catch_sigpipe(int sig)
{
	msg_err("caught SIGPIPE.");
	if (dt_state == DT_STATE_STARTED) {
		dtrace_terminate();
	}
}

int
dtrace_start(enum dt_script type, dev_t device)
{
	int 	fd[2];
	int 	pid;
	char 	script[BUF_SIZE];
	char 	switchrate[LINE_LEN];

	/* check if tracing already */
	if (dt_state == DT_STATE_STARTED) {
		return (FALSE);
	}

	/* catch sigpipe */
	if (sigset(SIGPIPE, dtrace_catch_sigpipe) == SIG_ERR) {
		msg_err("trap SIGPIPE failed. tracing not started.");
		return (FALSE);
	}

	/* use a pipe to a dtrace coprocess */
	if (pipe(fd) == -1) {
		msg_err("pipe failed. tracing not started.");
		return (FALSE);
	}

	/*
	 * Create DTrace coprocess
	 *
	 * I could write calls to libdtrace.so.1 directly, and I have
	 * written such consumers in the past, however I've choosen not to
	 * for three reasons,
	 *
	 *	a) you aren't supposed to. libdtrace is a private interface
	 *	   and isn't stable - it has changed already.
	 *	b) the performance difference isn't huge as the same DIF is
	 *	   run. it may be fair to argue that this way incurs pipe
	 *	   overheads and context switching, however these overheads
	 *	   have been reduced by using a slow switchrate.
	 *	c) the effort of reading libdtrace properly has already been
	 *	   coded into /usr/sbin/dtrace.
	 *
	 * A future version *may* use libdtrace, but I'd need to find a very
	 * good reason to.
	 */
	pid = fork();
	switch (pid) {

		case -1:
			msg_err("fork failed! tracing not started.");
			return (FALSE);

		case 0:
			close(fd[0]);

			/* set fds */
			if (fd[1] != STDOUT_FILENO)
				if (dup2(fd[1], STDOUT_FILENO) !=
				    STDOUT_FILENO)
					die_err("dup to stdout failed");

			/* ignore background tty write signal */
			(void) sigignore(SIGTTOU);

			/* put STDIN in raw mode */
			raw_stdin();

			/* merge dtrace script */
#ifdef _LP64
			sprintf(script, "%s%llu%s%d%s", dt_script0,
			    (u_longlong_t)device, dt_script1,
			    dt_heartbeat, dt_script2);
#else
			sprintf(script, "%s%lu%s%d%s", dt_script0,
			    (ulong_t)device, dt_script1,
			    dt_heartbeat, dt_script2);
#endif /* _LP64 */

			/* set switchrate */
			sprintf(switchrate, "switchrate=%dms", dt_heartbeat);

			/*
			 * Run DTrace
			 */
			execl("/usr/sbin/dtrace", "dtrace", "-x", switchrate,
			    "-qn", script, NULL);

			/* if we are here, somthing bad happened */
			die_err("exec failed");

		default:
			close(fd[1]);
			dt_dtrace_pid = pid;

			dt_dtrace_in = fdopen(fd[0], "r");
			/* BDG check errors */

			reset_stdin(0);
	}

	/* set global flag */
	dt_state = DT_STATE_STARTED;

	return (TRUE);
}

int
dtrace_isrunning()
{
	/* poll the dtrace process to see if it exists */
	if (dt_dtrace_pid > 3)
		return ((kill(dt_dtrace_pid, 0) == 0) ? TRUE : FALSE);
	else
		return (FALSE);
}

int
dtrace_close()
{
	/* close fd */
	if (fclose(dt_dtrace_in) == NULL)
		return (FALSE);

	return (TRUE);
}

int
dtrace_terminate()
{
	/* check if tracing first */
	if (dt_state != DT_STATE_STARTED) {
		return (FALSE);
	}

	/* first check child exists */
	if (dtrace_isrunning() == TRUE) {

		/* terminate coprocess */
		(void) kill(dt_dtrace_pid, SIGCONT);
		(void) kill(dt_dtrace_pid, SIGTERM);

		/*
		 * waitpid(dt_dtrace_pid, stat_loc, WNOHANG);
		 * BDG - waitpid does something odd to GTK. investigate.
		 */
	}

	/* set global flag */
	dt_state = DT_STATE_STOPPED;

	return (dtrace_close());
}

int
dtrace_read(char *buf)
{
	int	num;

	/* check there is data to read */
	ioctl(fileno(dt_dtrace_in), FIONREAD, &num);

	if (!num)
		return (0);	/* 0 is deliberate */

	/* clear buffer */
	bzero(buf, BUF_SIZE);

	/* get data */
	if (fgets(buf, num > BUF_SIZE ? BUF_SIZE : num, dt_dtrace_in) == NULL)
		return (0);	/* 0 is deliberate */

	/* return number of bytes */
	return (num);
}

int
dtrace_checkpulse(int delay)
{
	/* occasionally check dtrace's pulse */
	if (dt_dtrace_check++ > 1000 / delay) {

		dt_dtrace_check = 0;

		if (dtrace_isrunning() == FALSE) {

			/* the DTrace process stopped */
			dt_state = DT_STATE_STOPPED;
			dtrace_close();

			return (FALSE);
		}
	}

	return (TRUE);
}
